//
// Code based on helloflops3 from lotsofcores.com
// 
//
// A simple example that gets lots of Flops (Floating Point Operations) on 
// Intel(r) Xeon Phi(tm) co-processors using openmp to scale
//

// The work is scaled up in proportion to the number of threads




#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <omp.h>
#include <sys/time.h>

// dtime 
// 
// returns the current wall clock time
//
double dtime()
{
    double tseconds = 0.0;
    struct timeval mytime;
    gettimeofday(&mytime,(struct timezone*)0);
    tseconds = (double)(mytime.tv_sec + mytime.tv_usec*1.0e-6);
    return( tseconds );
}

#define FLOPS_ARRAY_SIZE (1024*1024*1024) 
#define MAXFLOPS_ITERS 100000000
#define LOOP_COUNT 128

// number of float pt ops per calculation
#define FLOPSPERCALC 2     
#define BYTESPERFLOAT 4
#define BYTESPERDOUBLE 8

// TODO:change these to try double precision AFTER vectorization
#define REAL float
#define BYTESPERREAL (4)

#define BYTESPERCALC (3*BYTESPERREAL)  //3 numbers * 4bytes per REAL

// define some arrays - 
// make sure they are 64 byte aligned
// for best cache access 

REAL *fa;
REAL *fb;


//
// Main program - pedal to the metal...calculate using tons o'flops!
// 
int main(int argc, char *argv[] ) 
{
    int i,j,k;
    int numthreads=1;
    double tstart, tstop, ttime;
    double gflops = 0.0;
    REAL a=1.1;
    float arithmetic_intensity = (float)FLOPSPERCALC/BYTESPERCALC;

    // Start some threads to get the number of threads from 
    //OMP_NUM_THREADS environment variable


    //TODO: Uncomment to get some threads

    /*
#pragma omp parallel
#pragma omp master
    numthreads = omp_get_num_threads();

    */


    printf("Initializing\r\n");

    
    fa = (REAL*)malloc(FLOPS_ARRAY_SIZE*sizeof(REAL));
    fb = (REAL*)malloc(FLOPS_ARRAY_SIZE*sizeof(REAL));
    


    //Split up the work of initializing the array --- not timing this
    //TODO: uncomment pragma
    /*
#pragma omp parallel for
    */
    for(i=0; i<FLOPS_ARRAY_SIZE; i++)
    {
      //initialization
      fa[i] = (REAL)i + 0.1;
      fb[i] = (REAL)i + 0.2;
     }	
    
    printf("Starting Compute on %d threads\r\n",numthreads);

    tstart = dtime();
	
    // scale the calculation across threads requested 
    // need to set environment variables OMP_NUM_THREADS and KMP_AFFINITY



    //TODO: uncomment pragma
    /*
#pragma omp parallel for private(j,k)
    */
    for (i=0; i<numthreads; i++)
      {
	// each thread will work it's own array section
	// calc offset into the right section
	  
	    int offset = i*LOOP_COUNT;
	
	
	
	// loop many times to get lots of calculations
	for(j=0; j<MAXFLOPS_ITERS; j++)  
	  {
	    // scale 1st array and add in the 2nd array 
	    for(k=0; k<LOOP_COUNT; k++)  
	      {
		
		//TODO: comment out the line below
		fa[k] = a * fa[k] + fb[k];
		
		//TODO: replace with line containing offset
		// so each thread works on a different part of array
		/*
		fa[k+offset] = a * fa[k+offset] + fb[k+offset];
		*/
		
	      }
	  }
      }

    tstop = dtime();
    // # of gigaflops we just calculated  
    gflops = (double)( 1.0e-9*LOOP_COUNT*numthreads*
		       MAXFLOPS_ITERS*FLOPSPERCALC);    
    
    
    //elasped time
    ttime = tstop - tstart;
    //
    // Print the results
    //
    if ((ttime) > 0.0)
    {
      printf("Threads = %d GFlops = %10.3lf, Secs = %10.3lf, GFlops per sec = %10.3lf\r\n",                numthreads,   gflops, ttime, gflops/ttime);
      printf("Bandwidth = %g GB/sec\n",gflops/arithmetic_intensity);
      printf("Arithmetic intensity=%g B/flop\n",arithmetic_intensity);

    }
    return( 0 );
}

// TRY: switching float -> double
